<?php

namespace Drupal\cl_block\Form;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Component\ComponentMetadata;
use Drupal\Component\Render\MarkupInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Block\BlockManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure Component Libraries: Blocks settings for this site.
 */
class SettingsForm extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function __construct(ConfigFactoryInterface $config_factory, private readonly ComponentDiscovery $componentDiscovery, private readonly BlockManagerInterface $blockManager) {
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $component_discovery = $container->get(ComponentDiscovery::class);
    assert($component_discovery instanceof ComponentDiscovery);
    return new static(
      $container->get('config.factory'),
      $component_discovery,
      $container->get('plugin.manager.block')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cl_block_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['cl_block.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $settings = $this->config('cl_block.settings')->get();
    // I am unsure why we need this, but the component discovery is not
    // initialized in ajax callbacks sometimes.
    $discovery = isset($this->componentDiscovery) ? $this->componentDiscovery : \Drupal::service(ComponentDiscovery::class);
    $message = $this->t('Set a combination of filters to control the list of components that will become blocks.');
    static::buildSettingsForm($form, $form_state, $discovery, $settings, [], $message);
    return parent::buildForm($form, $form_state);
  }

  /**
   * Builds the form.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param \Drupal\cl_components\Component\ComponentDiscovery $discovery
   *   The discovery.
   * @param array $settings
   *   Settings array.
   * @param \Drupal\Component\Render\MarkupInterface $message
   *   The info message.
   */
  public static function buildSettingsForm(array &$form, FormStateInterface $form_state, ComponentDiscovery $discovery, array $settings, array $parents, MarkupInterface $message): void {
    // Check the user input for the existing values.
    $values = $form_state->getValues();
    $raw_input = $form_state->getUserInput();
    $types = NestedArray::getValue($values, [...$parents, 'types'])
      ?? NestedArray::getValue($raw_input, [...$parents, 'types'])
      ?? $settings['types']
      ?? [];
    $statuses = NestedArray::getValue($values, [...$parents, 'statuses'])
      ?? NestedArray::getValue($raw_input, [...$parents, 'statuses'])
      ?? $settings['statuses']
      ?? [];
    $components = $discovery->findAllWithFilters([], [], $types, $statuses);
    $options = array_map(static fn (Component $component) => $component->getMetadata()->getName(), $components);
    $form['filters'] = [
      '#type' => 'fieldset',
      '#title' => t('Filters'),
    ];
    $form['filters']['info'] = [
      '#type' => 'html_tag',
      '#tag' => 'p',
      '#value' => $message,
    ];
    $form['filters']['types'] = [
      '#type' => 'checkboxes',
      '#title' => t('Component Types'),
      '#description' => t('Only components of these types will show in the component list.'),
      '#options' => [
        ComponentMetadata::COMPONENT_TYPE_ATOM => t('Atom'),
        ComponentMetadata::COMPONENT_TYPE_MOLECULE => t('Molecule'),
        ComponentMetadata::COMPONENT_TYPE_ORGANISM => t('Organism'),
      ],
      '#default_value' => $settings['types'],
      '#ajax' => [
        'wrapper' => 'refinement-wrapper',
        'effect' => 'fade',
        'callback' => [static::class, 'onChangeAjaxCallback'],
      ],
    ];
    $form['filters']['statuses'] = [
      '#type' => 'checkboxes',
      '#title' => t('Component Status'),
      '#description' => t('Only components of these status will show in the component list.'),
      '#options' => [
        ComponentMetadata::COMPONENT_STATUS_READY => t('Ready'),
        ComponentMetadata::COMPONENT_STATUS_BETA => t('Beta'),
        ComponentMetadata::COMPONENT_STATUS_WIP => t('Work in progress'),
        ComponentMetadata::COMPONENT_STATUS_DEPRECATED => t('Deprecated'),
      ],
      '#default_value' => $settings['statuses'],
      '#ajax' => [
        'wrapper' => 'refinement-wrapper',
        'effect' => 'fade',
        'callback' => [static::class, 'onChangeAjaxCallback'],
      ],
    ];
    $form['filters']['refine'] = [
      '#type' => 'details',
      '#open' => TRUE,
      '#title' => t('Refine selection'),
      '#prefix' => '<div id="refinement-wrapper">',
      '#suffix' => '</div>',
    ];
    $form['filters']['refine']['forbidden'] = [
      '#type' => 'checkboxes',
      '#title' => t('Forbidden components'),
      '#description' => t('When listing the blocks for CL Components, all components will be displayed except the components selected here. This is used to limit the components available for a given content type.'),
      '#options' => $options,
      '#default_value' => $settings['forbidden'],
      '#attributes' => ['class' => ['long-checkboxes-list']],
    ];
    $form['filters']['refine']['allowed'] = [
      '#type' => 'checkboxes',
      '#title' => t('Allowed components'),
      '#description' => t('When listing the blocks for CL Components, only the components selected here will be displayed as an option. <strong>NOTE:</strong> "forbidden" components will not be shown even when allowed here.'),
      '#options' => $options,
      '#default_value' => $settings['allowed'],
      '#attributes' => ['class' => ['long-checkboxes-list']],
    ];
    $form['filters']['#attached']['library'][] = 'cl_block/settings';
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $forbidden = array_keys(array_filter($form_state->getValue('forbidden')));
    $allowed = array_keys(array_filter($form_state->getValue('allowed')));
    $both_lists = array_intersect($forbidden, $allowed);
    if (!empty($both_lists)) {
      $message = $this->t('The following components are added to both forbidden and allow lists. Please review: %components', ['%components' => implode(', ', $both_lists)]);
      $form_state->setErrorByName('forbidden', $message);
      $form_state->setErrorByName('allowed', $message);
    }
    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $types = array_keys(array_filter($form_state->getValue('types')));
    $statuses = array_keys(array_filter($form_state->getValue('statuses')));
    $forbidden = array_keys(array_filter($form_state->getValue('forbidden')));
    $allowed = array_keys(array_filter($form_state->getValue('allowed')));
    $this->config('cl_block.settings')
      ->set('types', $types)
      ->set('statuses', $statuses)
      ->set('forbidden', $forbidden)
      ->set('allowed', $allowed)
      ->save();
    // Refresh the list of blocks after saving settings.
    $manager = isset($this->blockManager) ? $this->blockManager : \Drupal::service('plugin.manager.block');
    $manager->clearCachedDefinitions();
    parent::submitForm($form, $form_state);

  }

  /**
   * Ajax callback to regenerate the list of components.
   */
  public static function onChangeAjaxCallback(array &$element, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    // Pop the checkbox value and the "statuses" or "types" from the parents.
    // This should leave us with the parents to the "filters".
    $parents = array_slice($triggering_element['#array_parents'], 0, -2);
    return NestedArray::getValue($element, [...$parents, 'refine']);
  }

}
